# PAPI BasicAuth functions
# Copyright (c) 2001-2007, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.
#
# These functions implement PAPI user authentication and credential data 
# provision using a simple hash-tied Berkeley DB database.
#

package PAPI::BasicAuth;
use strict;
use DB_File;

# This function is intended to be hooked at the authenticationHook in a
# PAPI AuthServer. It uses the connection variables 'username' and 
# 'password' to validate them against the DB database defined by the
# configuration variable 'basicAuthDB'.
#
sub VerifyUser {
   my $con = shift;
   my $cfg = shift;
   my %badb;
   my $key = "user\:\:$$con{username}";
   my ($passwd, $alt, $grouplist, $ds, $galt);
   if (!exists $$cfg{'basicAuthDB'} or $$cfg{'basicAuthDB'} eq "") {
      $$con{'PAPIerror'} = "Undefined BasicAuth database";
      return undef;
   }
   if (!tie %badb, 'DB_File', $$cfg{'basicAuthDB'}, O_RDONLY, 0640) {
      $$con{'PAPIerror'} = "Error opening BasicAuth database: $!";
      return undef;
   }
   if (!exists $badb{$key}) { 
      $$con{'PAPIerror'} = "Unknown user";
      untie %badb;
      return undef;
   }
   ($passwd, $alt, $grouplist, $ds) = split /::/, $badb{$key};
   if (crypt ($$con{'password'}, $passwd) ne $passwd) {
      $$con{'PAPIerror'} = "Invalid password";
      untie %badb;
      return undef;
   }
   if ($alt eq "") { $alt = $$con{'username'}; }

# Look for alternate names in the groups the user belongs to. Set 'PAPIgid'
#
   for my $group (split /\s*,\s*/,$grouplist) {
      $key = "group\:\:$group";
      ($galt, $ds) = split /::/, $badb{$key};
      if ($galt ne "") {
         $$con{'PAPIgid'} = $galt;
         last;
      }
   }
   untie %badb;
   return $alt;
}


# This function is intended to be hooked at the credentialHook in a PAPI
# AuthServer. It uses the connection variable 'username' to return a list
# of site definitions the user can connect to. It returns data for the sites
# included in the user entry itself and in the entries for those group the
# user belongs to. The Berkeley DB database to be accessed must be defined
# by the configuration variable 'basicAuthDB'.
# If the configuration variable 'useDefaultSite' exists and is set to a
# non-zero value, it also includes an empty site definition, so the
# AuthServer appends a new site using the default values it is applying.
# 
sub UserCredentials {
   my $con = shift;
   my $cfg = shift;
   my %badb;
   my $key = "user\:\:$$con{username}";
   my @retar = ();
   my @sites = ();
   my ($dp, $da, $grouplist, $sitelist, $gsitelist, $six);
   my ($desc, $poa, $authURI, $ttl, $service, $location, $accessURI, $assertion);
   if (!exists $$cfg{'basicAuthDB'} or $$cfg{'basicAuthDB'} eq "") {
      $$con{'PAPIerror'} = "Undefined BasicAuth database";
      return \@retar;
   }
   if (!tie %badb, 'DB_File', $$cfg{'basicAuthDB'}, O_RDONLY, 0640) {
      $$con{'PAPIerror'} = "Error opening BasicAuth database: $!";
      return \@retar;
   }
   if (!exists $badb{$key}) {
      $$con{'PAPIerror'} = "Unknow user";
      untie %badb;
      return \@retar;
   }
   ($dp, $da, $grouplist, $sitelist) = split /::/, $badb{$key};
 
# Build a list of site identifiers, using first user's own references and
# then group references. Remove duplicate identifiers
   for my $group (split /\s*,\s*/,$grouplist) {
      $key = "group\:\:$group";
      ($da, $gsitelist) = split /::/, $badb{$key};
      if ($sitelist eq "") { $sitelist = $gsitelist; }
      else { $sitelist .= ",".$gsitelist; }
   }
   my @tsites = split /\s*,\s*/,$sitelist;
   for my $tsite (@tsites) {
      for ($six=0; $six<=$#sites; $six++) { last if ($sites[$six] eq $tsite); }
      push @sites, $tsite if ($six > $#sites);
   }
#
# Build the list of hashes from the site data stored into the database
   for my $site (@sites) {
      my $sitedef = {};
      $key = "site\:\:$site";
      ($desc, $poa, $authURI, $ttl, $service, $location, $accessURI, $assertion) = split /::/, $badb{$key};
      if ($desc ne "") { $sitedef->{'desc'} = $desc; }
      if ($poa ne "") { $sitedef->{'poa'} = $poa; }
      if ($authURI ne "") { $sitedef->{'authURI'} = $authURI; }
      if ($ttl ne "") { $sitedef->{'ttl'} = $ttl; }
      if ($service ne "") { $sitedef->{'service'} = $service; }
      if ($location ne "") { $sitedef->{'location'} = $location; }
      if ($accessURI ne "") { $sitedef->{'accessURI'} = $accessURI; }
      if ($assertion ne "") { $sitedef->{'assertion'} = $assertion; }
      push @retar, $sitedef;
   }
#
# Add default site if configured
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      push @retar, {};
   }
#
# Set error if no sites found
   $$con{'PAPIerror'} = "No suitable site found" if ($#retar < 0);
   untie %badb;
   return \@retar;
}


# This function is intended to be hooked at the credentialHook in a PAPI
# AuthServer. It returns a site list consisting of the sites related to
# the groups whose identifiers are included into the configuration
# variable 'groupList'. The list has to be comma-separated.
# If the configuration variable 'useDefaultSite' exists and is set to a
# non-zero value, it also includes an empty site definition, so the
# AuthServer appends a new site using the default values it is applying.
#
sub GroupList {
   my $con = shift;
   my $cfg = shift;
   my %badb;
   my @retar = ();
   my @sites = ();
   my ($key, $dp, $da, $sitelist, $gsitelist, $six);
   my ($desc, $poa, $authURI, $ttl, $service, $location, $accessURI, $assertion);
   my ($galt, $ds);
   if (!exists $$cfg{'basicAuthDB'} or $$cfg{'basicAuthDB'} eq "") {
      $$con{'PAPIerror'} = "Undefined BasicAuth database";
      return \@retar;
   }
   if (!tie %badb, 'DB_File', $$cfg{'basicAuthDB'}, O_RDONLY, 0640) {
      $$con{'PAPIerror'} = "Error opening BasicAuth database: $!";
      return \@retar;
   }
 
# Build a list of site identifiers. Remove duplicate identifiers
   for my $group (split /\s*,\s*/, $$cfg{'groupList'}) {
      $key = "group\:\:$group";
      ($da, $gsitelist) = split /::/, $badb{$key};
      if ($sitelist eq "") { $sitelist = $gsitelist; }
      else { $sitelist .= ",".$gsitelist; }
   }
   my @tsites = split /\s*,\s*/,$sitelist;
   for my $tsite (@tsites) {
      for ($six=0; $six<=$#sites; $six++) { last if ($sites[$six] eq $tsite); }
      push @sites, $tsite if ($six > $#sites);
   }
#
# Build the list of hashes from the site data stored into the database
   for my $site (@sites) {
      my $sitedef = {};
      $key = "site\:\:$site";
      ($desc, $poa, $authURI, $ttl, $service, $location, $accessURI, $assertion) = split /::/, $badb{$key};
      if ($desc ne "") { $sitedef->{'desc'} = $desc; }
      if ($poa ne "") { $sitedef->{'poa'} = $poa; }
      if ($authURI ne "") { $sitedef->{'authURI'} = $authURI; }
      if ($ttl ne "") { $sitedef->{'ttl'} = $ttl; }
      if ($service ne "") { $sitedef->{'service'} = $service; }
      if ($location ne "") { $sitedef->{'location'} = $location; }
      if ($accessURI ne "") { $sitedef->{'accessURI'} = $accessURI; }
      if ($assertion ne "") { $sitedef->{'assertion'} = $assertion; }
      push @retar, $sitedef;
   }
#
# Add default site if configured
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      push @retar, {};
   }
# Look for alternate names in the groups the user belongs to. Set 'PAPIgid'
#
   for my $group (split /\s*,\s*/,$$cfg{'groupList'}) {
      $key = "group\:\:$group";
      ($galt, $ds) = split /::/, $badb{$key};
      if ($galt ne "") {
         $$con{'PAPIgid'} = $galt;
         last;
      }
   }
#
# Set error if no sites found
   $$con{'PAPIerror'} = "No suitable site found" if ($#retar < 0);
   untie %badb;
   return \@retar;
}


# This function is intended to be hooked at the credentialHook in a PAPI
# AuthServer. It returns a site list consisting of all the sites included
# into the database.
# If the configuration variable 'useDefaultSite' exists and is set to a
# non-zero value, it also includes an empty site definition, so the
# AuthServer appends a new site using the default values it is applying.
#
sub AllSites {
   my $con = shift;
   my $cfg = shift;
   my %badb;
   my @retar = ();
   my ($desc, $poa, $authURI, $ttl, $service, $location, $accessURI, $assertion);
   if (!exists $$cfg{'basicAuthDB'} or $$cfg{'basicAuthDB'} eq "") {
      $$con{'PAPIerror'} = "Undefined BasicAuth database";
      return \@retar;
   }
   if (!tie %badb, 'DB_File', $$cfg{'basicAuthDB'}, O_RDONLY, 0640) {
      $$con{'PAPIerror'} = "Error opening BasicAuth database: $!";
      return \@retar;
   }
#
# Explore the DB-tied hash looking for site definitions
   for my $k (keys %badb) {
      if ($k =~ /^site::/) {
         my $sitedef = {};
         ($desc, $poa, $authURI, $ttl, $service, $location, $accessURI, $assertion) = split /::/, $badb{$k};
         if ($desc ne "") { $sitedef->{'desc'} = $desc; }
         if ($poa ne "") { $sitedef->{'poa'} = $poa; }
         if ($authURI ne "") { $sitedef->{'authURI'} = $authURI; }
         if ($ttl ne "") { $sitedef->{'ttl'} = $ttl; }
         if ($service ne "") { $sitedef->{'service'} = $service; }
         if ($location ne "") { $sitedef->{'location'} = $location; }
         if ($accessURI ne "") { $sitedef->{'accessURI'} = $accessURI; }
         if ($assertion ne "") { $sitedef->{'assertion'} = $assertion; }
         push @retar, $sitedef;
      }
   }
#
# Add default site if configured
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      push @retar, {};
   }
#
# Set error if no sites found
   $$con{'PAPIerror'} = "No suitable site found" if ($#retar < 0);
   untie %badb;
   return \@retar;
}


# This function is intended to be hooked at the credentialHook in a PAPI
# AuthServer. It just returns an empty site list, so the default values
# configured for the AuthServer are used. To be used as a stub when no other
# more sophisticated credentials assignment methods are requested.
#
sub DefCredentials {
   my @retar = ();
   push @retar, {};
   return \@retar;
}


# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to look
# for a defined site in the database acceptable to the user. The user
# id is read from the connection variable 'PAPIuid' (probably derived by the
# AuthServer from a authentication coookie). If a suitable site is found, it
# returns the server-wide assertion format and the time-to-live defined for
# the site. If no site is found matching the requesting URL, it sets the
# 'PAPIerror' variable and a negative value is returned.
#
sub UserAttributes {
   my $con = shift;
   my $cfg = shift;
   my $userId = "user\:\:$$con{PAPIuid}";
   my @retar = (-1, undef);
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};
   my %badb = ();
   my @sites = ();
   my ($dp, $da, $grouplist, $sitelist, $gsitelist, $six);
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   if (!exists $$cfg{'basicAuthDB'} or $$cfg{'basicAuthDB'} eq "") {
      $$con{'PAPIerror'} = "Undefined BasicAuth database";
      return @retar;
   }
   if (!tie %badb, 'DB_File', $$cfg{'basicAuthDB'}, O_RDONLY, 0640) {
      $$con{'PAPIerror'} = "Error opening BasicAuth database: $!";
      return @retar;
   }
   if (!exists $badb{$userId}) {
      $$con{'PAPIerror'} = "Unknow user $userId for PoA URL $poaurl";
      untie %badb;
      return @retar;
   }
   my ($dp, $da, $grouplist, $sitelist) = split /::/, $badb{$userId};
#
# Build a list of site identifiers, using first user's own references and
# then group references. Remove duplicate identifiers
   for my $group (split /\s*,\s*/,$grouplist) {
      my $key = "group\:\:$group";
      ($da, $gsitelist) = split /::/, $badb{$key};
      if ($sitelist eq "") { $sitelist = $gsitelist; }
      else { $sitelist .= ",".$gsitelist; }
   }
   my @tsites = split /\s*,\s*/,$sitelist;
   for my $tsite (@tsites) {
      for ($six=0; $six<=$#sites; $six++) { last if ($sites[$six] eq $tsite); }
      push @sites, $tsite if ($six > $#sites);
   }
#
# Explore the sites in the site list. Keep the longest match.
# Try frist with the original PoA
   my $len = 0;
   my $rettl = undef;
   my $reassertion = undef;
   if (defined $opoa) {
     for my $site (@sites) {
       my $key = "site\:\:$site";
       my ($dsc, $poa, $authURI, $ttl, $srv, $loc, $accessURI, $assertion) = split /::/, $badb{$key};
       my $siu = $poa.$loc;
       if ($opoa =~ m(^$siu) && length $siu > $len) { $rettl = $ttl; $reassertion = $assertion; }
     }
   }
   if (!defined $rettl) {
     for my $site (@sites) {
       my $key = "site\:\:$site";
       my ($dsc, $poa, $authURI, $ttl, $srv, $loc, $accessURI, $assertion) = split /::/, $badb{$key};
       my $siu = $poa.$loc;
       if ($poaurl =~ m(^$siu) && length $siu > $len) { $rettl = $ttl; $reassertion = $assertion; }
     }
   }
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
      if (($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) && length $siu > $len) {
         $rettl = $$cfg{'defTimeToLive'};
         $reassertion = undef;   # Will be assigned to later 
      }
   }
   if (!defined $rettl) {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";
      return @retar;
   }
   $retar[0] = $rettl;
   if (defined $reassertion && $reassertion ne "") { $retar[1] = $reassertion; }
   elsif (exists $$cfg{'defAssertion'}) { $retar[1] = $$cfg{'defAssertion'}; }
   else { $retar[1] = '<papi var="PAPIuid"/>-<papi var="PAPIgid"/>'; }
   return @retar;
}


# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to look
# for a defined site in the database acceptable to a list of groups,
# identified by the configuration variable 'groupList'. If a suitable site
# is found, it returns the server-wide assertion format and the time-to-live
# defined for the site. If no site is found matching the requesting URL, it
# sets the 'PAPIerror' variable and a negative value is returned.
#
sub GroupListAttributes {
   my $con = shift;
   my $cfg = shift;
   my @retar = (-1, undef);
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};
   my %badb = ();
   my @sites = ();
   my ($da, $sitelist, $gsitelist, $six);
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   if (!exists $$cfg{'basicAuthDB'} or $$cfg{'basicAuthDB'} eq "") {
      $$con{'PAPIerror'} = "Undefined BasicAuth database";
      return @retar;
   }
   if (!tie %badb, 'DB_File', $$cfg{'basicAuthDB'}, O_RDONLY, 0640) {
      $$con{'PAPIerror'} = "Error opening BasicAuth database: $!";
      return @retar;
   }
#
# Build a list of site identifiers. Remove duplicate identifiers
   for my $group (split /\s*,\s*/, $$cfg{'groupList'}) {
      my $key = "group\:\:$group";
      ($da, $gsitelist) = split /::/, $badb{$key};
      if ($sitelist eq "") { $sitelist = $gsitelist; }
      else { $sitelist .= ",".$gsitelist; }
   }
   my @tsites = split /\s*,\s*/,$sitelist;
   for my $tsite (@tsites) {
      for ($six=0; $six<=$#sites; $six++) { last if ($sites[$six] eq $tsite); }
      push @sites, $tsite if ($six > $#sites);
   }
#   
# Explore the sites in the site list. Keep the longest match.
# Try frist with the original PoA
   my $len = 0;
   my $rettl = undef;
   my $reassertion = undef;
   if (defined $opoa) {
     for my $site (@sites) {
       my $key = "site\:\:$site";
       my ($dsc, $poa, $authURI, $ttl, $srv, $loc, $accessURI, $assertion) = split /::/, $badb{$key};
       my $siu = $poa.$loc;
       if ($opoa =~ m(^$siu) && length $siu > $len) { $rettl = $ttl; $reassertion = $assertion; }
     }
   }
   if (!defined $rettl) {
     for my $site (@sites) {
       my $key = "site\:\:$site";
       my ($dsc, $poa, $authURI, $ttl, $srv, $loc, $accessURI, $assertion) = split /::/, $badb{$key};
       my $siu = $poa.$loc;
       if ($poaurl =~ m(^$siu) && length $siu > $len) { $rettl = $ttl; $reassertion = $assertion; }
     }
   }
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
      if (($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) && length $siu > $len) {
          $rettl = $$cfg{'defTimeToLive'};
         $reassertion = undef;   # Will be assigned to later 
      }
   }
   if (!defined $rettl) {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";
      return @retar;
   }
   $retar[0] = $rettl;
   if (defined $reassertion && $reassertion ne "") { $retar[1] = $reassertion; }
   elsif (exists $$cfg{'defAssertion'}) { $retar[1] = $$cfg{'defAssertion'}; }
   else { $retar[1] = '<papi var="PAPIuid"/>-<papi var="PAPIgid"/>'; }
   return @retar;
}


# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to look
# for a defined site in the database. If a suitable site is found, it
# returns the server-wide assertion format and the time-to-live defined
# for the site. If no site is found matching the requesting URL, it sets
# the 'PAPIerror' variable and a negative value is returned.
#
sub AllSiteAttributes {
   my $con = shift;
   my $cfg = shift;
   my @retar = (-1, undef);
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   my %badb = ();
   if (exists $$cfg{'basicAuthDB'} and $$cfg{'basicAuthDB'} ne "") {
      if (!tie %badb, 'DB_File', $$cfg{'basicAuthDB'}, O_RDONLY, 0640) {
         $$con{'PAPIerror'} = "Error opening BasicAuth database: $!";
      }
   }
#   
# Explore the sites in the site list. Keep the longest match.
# Try frist with the original PoA
   my $len = 0;
   my $rettl = undef;
   my $reassertion = undef;
   if (defined $opoa) {
     for my $k (keys %badb) {
       if ($k =~ /^site::/) {
         my ($dsc, $poa, $authURI, $ttl, $srv, $loc, $accessURI, $assertion) = split /::/, $badb{$k};
         my $siu = $poa.$loc;
         if ($opoa =~ m(^$siu) && length $siu > $len) { $rettl = $ttl; $reassertion = $assertion; }
       }
     }
   }
   if (!defined $rettl) {
     for my $k (keys %badb) {
       if ($k =~ /^site::/) {
         my ($dsc, $poa, $authURI, $ttl, $srv, $loc, $accessURI, $assertion) = split /::/, $badb{$k};
         my $siu = $poa.$loc;
         if ($poaurl =~ m(^$siu) && length $siu > $len) { $rettl = $ttl; $reassertion = $assertion; }
       }
     }
   }
   if (exists $$cfg{'useDefaultSite'} and $$cfg{'useDefaultSite'} != 0) {
      my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
      if (($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) && length $siu > $len) {
         $rettl = $$cfg{'defTimeToLive'};
         $reassertion = undef;   # Will be assigned to later 
      }
   }
   if (!defined $rettl) {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";
      return @retar;
   }
   $retar[0] = $rettl;
   if (defined $reassertion && $reassertion ne "") { $retar[1] = $reassertion; }
   elsif (exists $$cfg{'defAssertion'}) { $retar[1] = $$cfg{'defAssertion'}; }
   else { $retar[1] = '<papi var="PAPIuid"/>-<papi var="PAPIgid"/>'; }
   return @retar;
}


# This function is intended to be hooked at the attrRequestHook in a
# PAPI AuthServer. It uses the connection variable 'PAPIPOAURL' to match
# against the default site defined for the AuthServer. To be used as a stub
# when no other more sophisticated methods are requested.
#
sub DefAttributes {
   my $con = shift;
   my $cfg = shift;
   my @retar = (-1, undef);
   my $poaurl = $$con{'PAPIPOAURL'};
   my $opoa = $$con{'PAPIOPOA'};
   if (!defined $poaurl) {
      $$con{'PAPIerror'} = "Undefined PoA referencing URL";
      return @retar;
   }
   my $siu = $$cfg{'defPoA'}.$$cfg{'defLocation'};
   if ($poaurl =~ m(^$siu) || $opoa =~ m(^$siu)) {
      $retar[0] = $$cfg{'defTimeToLive'};
      if (exists $$cfg{'defAssertion'}) { $retar[1] = $$cfg{'defAssertion'}; }
      else { $retar[1] = '<papi var="PAPIuid"/>-<papi var="PAPIgid"/>'; }
   }
   else {
      $$con{'PAPIerror'}="No suitable site found for requesting PoA at $poaurl";
   }
   return @retar;
}

1;

